import os
from osgeo import gdal
from .transform import ds_name, context_file
from ._const import CREATION, CONFIG


__all__ = ['resample']


def resample(ds, out_path, **kwargs):
    ds, ras = ds_name(ds)
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    resample_alg = kwargs.pop('resampleAlg', gdal.GRA_Average)
    option = gdal.WarpOptions(multithread=True, options=CONFIG,
                              creationOptions=CREATION, **kwargs,
                              resampleAlg=resample_alg)
    gdal.Warp(out_file, ds, options=option)

    return out_file



import os
import json
import numpy as np
from osgeo import gdal

__all__ = ['context_file', 'ds_name']


def rep_file(cache_dir, filename):
    prefix, extension = os.path.splitext(os.path.basename(filename))
    file_path = os.path.join(cache_dir, prefix + extension)
    if os.path.exists(file_path):
        i = 1
        while True:
            file_path = os.path.join(
                cache_dir, prefix + '(' + str(i) + ')' + extension)
            if os.path.exists(file_path):
                i += 1
            else:
                break
    return file_path


def geo2imagexy(ds, x, y):
    trans = ds.GetGeoTransform()
    a = np.array([[trans[1], trans[2]], [trans[4], trans[5]]])
    b = np.array([x - trans[0], y - trans[3]])
    col, row = np.linalg.solve(a, b) - 0.5
    return int(round(col)), int(round(row))


def context_file(ras, out_path):
    ext = os.path.splitext(os.path.basename(out_path))[1]
    if ext != '.tif':
        if not os.path.isdir(out_path):
            os.makedirs(out_path)
        out_file = os.path.join(out_path, os.path.splitext(
            os.path.basename(ras))[0] + '.tif')
    else:
        out_file = out_path

    return out_file


def ds_name(ds):
    if isinstance(ds, str):
        ras = ds
        ds = gdal.Open(ras)
    else:
        ras = ds.GetDescription()

    return ds, ras


def get_extent(layer):
    r = layer.GetSpatialRef()
    crs = json.loads(r.ExportToPROJJSON())
    axis = [crs['coordinate_system']['axis'][i]['direction'].upper()
            for i in range(r.GetAxesCount())]
    order = ['EAST', 'WEST', 'NORTH', 'SOUTH']
    order_dict = dict(zip(order, range(4)))
    axis_code = np.array([order_dict[i] for i in axis])
    axis_sort = axis_code.argsort()
    data_map = [abs(i) - 1 for i in r.GetDataAxisToSRSAxisMapping()]
    extent = np.reshape(layer.GetExtent(), [-1, 2])[data_map]
    xy_extent = np.reshape(extent[axis_sort], [-1])
    return tuple(xy_extent)



from osgeo import gdal, osr
from ._const import CREATION


def zeros_tif(out_file, x_size, y_size, n_band,
              data_type, trans, srs, no_data=2):

    ds = gdal.GetDriverByName('GTiff').Create(
        out_file, x_size, y_size, n_band, data_type, CREATION)

    # fill with 0 and set no data
    band = ds.GetRasterBand(1)
    band.Fill(0)
    band.SetNoDataValue(no_data)

    # set geotransform
    ds.SetGeoTransform(tuple(trans))

    # set SpatialReference
    if isinstance(srs, osr.SpatialReference):
        inSpatialRef = srs
    else:
        inSpatialRef = osr.SpatialReference()
        inSpatialRef.ImportFromProj4(srs)
    ds.SetProjection(srs.ExportToWkt())

    band = None
    ds = None

    return out_file



from ._const import CREATION, CONFIG
from .bound import bound_raster, bound_layers
from .block_write import block_write
from .convert_uint8 import convert_uint8
from .polygonize import polygonize
from .shp_to_raster import shp_to_raster
from .projection import proj_shapefile
from .buffer import shp_buffer
from .split import split_shp
from .extract import extract
from .resample import resample
from .gdal_calc import Calc
from .grib_to_tif import grib_to_tif
from .mosaic import mosaic
from .map_calc import map_calc
from .transform import rep_file, geo2imagexy, context_file, ds_name


__all__ = ['bound', 'block_write', 'convert_uint8',
           'polygonize', 'shp_to_raster', 'projection',
           'buffer', 'split', 'extract', 'resample',
           'gdal_calc', 'grib_to_tif', 'mosaic', 'map_calc',
           'transform']



import os
import numpy as np
from osgeo import gdal, ogr, osr
from ._const import CREATION, CONFIG
from .transform import geo2imagexy, rep_file
from .block_write import block_write
from .create_tif import zeros_tif
from .projection import proj_shapefile


__all__ = ['extract']


# Now Read the large raster block by block, expecting to see no increase of
# memory as we loop through the blocks.
def map_burn(in_datas):
    in_datas[1].set_fill_value(0)
    burn_data = in_datas[1].filled()
    rect_data = np.ma.masked_where(np.logical_not(
        burn_data.astype(np.bool)), in_datas[0])
    return rect_data.filled()


def clip(ds, outLayer, no_data=None, rect_file=None,
         enlarge=10, save_cache=False, ext=None, new=True):

    # Open the data source and read in the extent
    t = ds.GetGeoTransform()
    x_min, x_max, y_min, y_max = outLayer.GetExtent()
    ulX, ulY = geo2imagexy(ds, x_min, y_min)
    lrX, lrY = geo2imagexy(ds, x_max, y_max)
    clip_range = [min(ulX, lrX), min(ulY, lrY),
                  abs(ulX - lrX) + 1, abs(ulY - lrY) + 1]
    ul_lon = t[0] + t[1] * clip_range[0] + t[2] * clip_range[1]
    ul_lat = t[3] + t[4] * clip_range[0] + t[5] * clip_range[1]
    lr_lon = t[0] + t[1] * (clip_range[0] + clip_range[2]) + \
        t[2] * (clip_range[1] + clip_range[3])
    lr_lat = t[3] + t[4] * (clip_range[0] + clip_range[2]) + \
        t[5] * (clip_range[1] + clip_range[3])
    bound = [min(ul_lon, lr_lon), min(ul_lat, lr_lat),
             max(ul_lon, lr_lon), max(ul_lat, lr_lat)]

    if save_cache:
        cache_dir = 'cache'
        if not os.path.exists(cache_dir):
            os.mkdir(cache_dir)

        if ext is None:
            poly_file = rep_file(cache_dir, 'poly.tif')
            burn_file = rep_file(cache_dir, 'burn.tif')
        elif new:
            poly_file = rep_file(cache_dir, str(ext) + '_poly.tif')
            burn_file = rep_file(cache_dir, str(ext) + '_burn.tif')
        else:
            poly_file = os.path.join(cache_dir, str(ext) + '_poly.tif')
            burn_file = os.path.join(cache_dir, str(ext) + '_burn.tif')
    else:
        poly_file = '/vsimem/_poly.tif'
        burn_file = '/vsimem/_burn.tif'

    # set no data
    if ds.GetRasterBand(1).GetNoDataValue() is not None:
        no_data = ds.GetRasterBand(1).GetNoDataValue()
    if no_data is None:
        raise(ValueError("no_data must be initialed"))

    # create temp bool in_poly tif
    has_old = os.path.exists(burn_file) and not new

    # clip with rectangle
    if rect_file is None:
        rect_file = '/vsimem/_rect.tif'
    option = gdal.WarpOptions(multithread=True, options=CONFIG,
                              creationOptions=CREATION, outputBounds=bound,
                              dstSRS=outLayer.GetSpatialRef(), dstNodata=no_data,
                              xRes=t[1], yRes=t[5], srcNodata=no_data,
                              resampleAlg=gdal.GRA_NearestNeighbour)
    rect = gdal.Warp(rect_file, ds, options=option)

    if not has_old:
        # set geotransform
        trans = list(rect.GetGeoTransform())
        trans[1] = trans[1] / enlarge
        trans[5] = trans[5] / enlarge

        # set SpatialReference
        srs = outLayer.GetSpatialRef()

        zeros_tif(poly_file, int(clip_range[2] * enlarge),
                  int(clip_range[3] * enlarge), 1,
                  gdal.GDT_Byte, trans, srs, no_data=2)
        poly_ds = gdal.Open(poly_file, gdal.GA_Update)

        # Rasterize
        gdal.RasterizeLayer(poly_ds, [1], outLayer, burn_values=[
                            1], options=['ALL_TOUCHED=TRUE'])
        poly_ds = None

        option = gdal.WarpOptions(multithread=True, options=CONFIG,
                                  creationOptions=CREATION,
                                  xRes=rect.GetGeoTransform()[1],
                                  yRes=rect.GetGeoTransform()[5],
                                  resampleAlg=gdal.GRA_Average,
                                  outputType=gdal.GDT_Float32)
        burn_ds = gdal.Warp(burn_file, poly_file, options=option)
        burn_ds.GetRasterBand(1).SetNoDataValue(0)

        # return bool matrix in polygon
        burn_band = burn_ds.GetRasterBand(1)
        burn_data = burn_band.ReadAsArray()
    else:
        burn_ds = gdal.Open(burn_file)
        burn_band = burn_ds.GetRasterBand(1)
        burn_data = burn_band.ReadAsArray()

    # change rect
    for c in range(1, rect.RasterCount + 1):
        rect_band = rect.GetRasterBand(c)
        if rect_band.GetNoDataValue() is None:
            rect_band.SetNoDataValue(no_data)
        block_write(rect, [rect_band, burn_band], rect_band, map_burn)

    burn_ds = None
    poly_ds = None

    return rect, burn_data


def extract_stat(ras, shp, PROJ=None, no_data=None, **kwargs):
    if isinstance(ras, str):
        ds = gdal.Open(ras)
    else:
        ds = ras

    # Filename of input OGR file
    driver = ogr.GetDriverByName("ESRI Shapefile")
    source_ds = driver.Open(shp)
    inLayer = source_ds.GetLayer()
    inLayerDefn = inLayer.GetLayerDefn()

    # input SpatialReference
    inSpatialRef = inLayer.GetSpatialRef()

    # output SpatialReference
    outSpatialRef = osr.SpatialReference()
    if ds.GetProjectionRef():
        prj_srs = ds.GetProjectionRef()
        outSpatialRef.ImportFromWkt(prj_srs)
    elif PROJ:
        outSpatialRef.ImportFromProj4(PROJ)
        ds.SetProjection(outSpatialRef.ExportToWkt())
    else:
        raise(ValueError("PROJ must be initialed"))

    # create the CoordinateTransformation
    if int(gdal.__version__[0]) >= 3:
        outSpatialRef.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    coordTrans = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)

    # initial output
    row = inLayer.GetFeatureCount()
    col = ds.RasterCount
    stat = np.full((row, col), np.nan)

    for r in range(row):
        inFeature = inLayer.GetFeature(r)
        # create the output layer
        outFile = '/vsimem/outline.shp'
        if os.path.exists(outFile):
            driver.DeleteDataSource(outFile)
        outDataSet = driver.CreateDataSource(outFile)
        outLayer = outDataSet.CreateLayer(outFile, outSpatialRef)

        # add fields
        for i in range(0, inLayerDefn.GetFieldCount()):
            fieldDefn = inLayerDefn.GetFieldDefn(i)
            outLayer.CreateField(fieldDefn)

        # get the output layer's feature definition
        outLayerDefn = outLayer.GetLayerDefn()

        # get the input geometry
        geom = inFeature.GetGeometryRef()
        # reproject the geometry
        geom.Transform(coordTrans)
        # create a new feature
        outFeature = ogr.Feature(outLayerDefn)
        # set the geometry and attribute
        outFeature.SetGeometry(geom)
        for i in range(0, outLayerDefn.GetFieldCount()):
            outFeature.SetField(outLayerDefn.GetFieldDefn(
                i).GetNameRef(), inFeature.GetField(i))
        # add the feature to the shapefile
        outLayer.CreateFeature(outFeature)

        if ds.GetRasterBand(1).GetNoDataValue() is not None:
            no_data = ds.GetRasterBand(1).GetNoDataValue()
        if no_data is None:
            raise(ValueError("no_data must be initialed"))
        rect = None
        rect, burn_data = clip(ds, outLayer, no_data=no_data, **kwargs)

        # iterate all bands
        for c in range(1, col + 1):
            rect_band = rect.GetRasterBand(c)
            rect_data = rect_band.ReadAsArray()
            no_data = rect_band.GetNoDataValue()
            select = (rect_data != no_data)
            try:
                stat[r, c - 1] = np.average(rect_data[select],
                                            weights=burn_data[select])
            except ZeroDivisionError:
                pass

        outFeature = None
        outDataSet = None
        outLayer = None
        inFeature = None

    ds = None
    return stat


def extract(ras, shp, PROJ=None, no_data=None, stat=False, **kwargs):
    if stat:
        return extract_stat(ras, shp, PROJ=PROJ, no_data=no_data, **kwargs)

    if isinstance(ras, str):
        ds = gdal.Open(ras)
    else:
        ds = ras

    # output SpatialReference
    outSpatialRef = osr.SpatialReference()
    if ds.GetProjectionRef():
        prj_srs = ds.GetProjectionRef()
        outSpatialRef.ImportFromWkt(prj_srs)
    elif PROJ:
        prj_srs = PROJ
        outSpatialRef.ImportFromProj4(prj_srs)
        ds.SetProjection(outSpatialRef.ExportToWkt())
    else:
        raise(ValueError("PROJ must be initialed"))

    out_shp = '/vsimem/outline.shp'
    proj_shapefile(shp, out_shp, out_proj=outSpatialRef)
    outDataSet = ogr.Open(out_shp)
    outLayer = outDataSet.GetLayer()

    # set no data
    if ds.GetRasterBand(1).GetNoDataValue() is not None:
        no_data = ds.GetRasterBand(1).GetNoDataValue()
    if no_data is None:
        raise(ValueError("no_data must be initialed"))

    # clip with shapefile
    rect, _ = clip(ds, outLayer, no_data=no_data, **kwargs)

    return rect.GetDescription()



import numpy as np
from osgeo import gdal, ogr, osr
from .transform import geo2imagexy, ds_name


__all__ = ['bound_raster', 'bound_layers']


def proj_bound(ds, bound, bound_srs):
    x_min, y_min, x_max, y_max = bound
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(min(x_min, x_max), min(y_min, y_max))
    ring.AddPoint(max(x_min, x_max), min(y_min, y_max))
    ring.AddPoint(max(x_min, x_max), max(y_min, y_max))
    ring.AddPoint(min(x_min, x_max), max(y_min, y_max))
    ring.AddPoint(min(x_min, x_max), min(y_min, y_max))

    # Create polygon
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)

    # output SpatialReference
    if isinstance(bound_srs, osr.SpatialReference):
        outSpatialRef = bound_srs
    else:
        outSpatialRef = osr.SpatialReference()
        outSpatialRef.ImportFromProj4(bound_srs)

    # raster spatial reference
    ras_srs = osr.SpatialReference()
    ras_srs.ImportFromWkt(ds.GetProjection())

    # create the CoordinateTransformation
    if int(gdal.__version__[0]) >= 3:
        outSpatialRef.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        ras_srs.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    trans = osr.CoordinateTransformation(outSpatialRef, ras_srs)
    trans_reverse = osr.CoordinateTransformation(ras_srs, outSpatialRef)

    # create a geometry from coordinates
    t = ds.GetGeoTransform()
    point = ogr.Geometry(ogr.wkbPoint)
    point_dx = ogr.Geometry(ogr.wkbPoint)
    point_dy = ogr.Geometry(ogr.wkbPoint)
    point.AddPoint(t[0] + ds.RasterXSize // 2 * t[1] + ds.RasterYSize // 2 *
                   t[2], t[3] + ds.RasterXSize // 2 * t[4] + ds.RasterYSize // 2 * t[5])
    point_dx.AddPoint(t[0] + (ds.RasterXSize // 2 + 1) * t[1] + ds.RasterYSize // 2 *
                      t[2], t[3] + (ds.RasterXSize // 2 + 1) * t[4] + ds.RasterYSize // 2 * t[5])
    point_dy.AddPoint(t[0] + ds.RasterXSize // 2 * t[1] + (ds.RasterYSize // 2 + 1) *
                      t[2], t[3] + ds.RasterXSize // 2 * t[4] + (ds.RasterYSize // 2 + 1) * t[5])
    point.Transform(trans_reverse)
    point_dx.Transform(trans_reverse)
    point_dy.Transform(trans_reverse)
    dx = abs(point.GetPoint()[0] - point_dx.GetPoint()[0])
    dy = abs(point.GetPoint()[1] - point_dy.GetPoint()[1])

    # density geom
    geom = ogr.CreateGeometryFromWkb(poly.ExportToWkb())
    geom.Segmentize(min(dx, dy) / 2)
    geom.Transform(trans)

    # get boundary
    win = np.array([geom.GetEnvelope()])
    bound = [win[:, 0:2].min(), win[:, 2:].min(),
             win[:, 0:2].max(), win[:, 2:].max()]
    return bound


def bound_raster(ds, bound, bound_srs="+proj=longlat +datum=WGS84 +ellps=WGS84"):
    ds, ras = ds_name(ds)
    t = ds.GetGeoTransform()
    x_min, y_min, x_max, y_max = proj_bound(ds, bound, bound_srs)
    ulX, ulY = geo2imagexy(ds, x_min, y_min)
    lrX, lrY = geo2imagexy(ds, x_max, y_max)
    clip_range = [min(ulX, lrX), min(ulY, lrY),
                  abs(ulX - lrX) + 1, abs(ulY - lrY) + 1]
    ul_lon = t[0] + t[1] * clip_range[0] + t[2] * clip_range[1]
    ul_lat = t[3] + t[4] * clip_range[0] + t[5] * clip_range[1]
    lr_lon = t[0] + t[1] * (clip_range[0] + clip_range[2]) + \
        t[2] * (clip_range[1] + clip_range[3])
    lr_lat = t[3] + t[4] * (clip_range[0] + clip_range[2]) + \
        t[5] * (clip_range[1] + clip_range[3])
    bound = [min(ul_lon, lr_lon), min(ul_lat, lr_lat),
             max(ul_lon, lr_lon), max(ul_lat, lr_lat)]

    # raster spatial reference as bound srs
    bound_srs = osr.SpatialReference()
    bound_srs.ImportFromWkt(ds.GetProjection())

    return bound, bound_srs.ExportToProj4()


def bound_layers(layers):
    # clip extent
    driver = ogr.GetDriverByName('ESRI Shapefile')
    if not isinstance(layers, list):
        layers = [layers]
    if isinstance(layers[0], str):
        shp_in = [driver.Open(lyr) for lyr in layers]
        layers = [s.GetLayer() for s in shp_in]
    win = np.array([lyr.GetExtent() for lyr in layers])
    bound = [win[:, 0:2].min(), win[:, 2:].min(),
             win[:, 0:2].max(), win[:, 2:].max()]
    bound_srs = layers[0].GetSpatialRef()

    return bound, bound_srs.ExportToProj4()



import numpy as np
from collections.abc import Iterable


__all__ = ['block_write']


def block_write(in_ds, in_bands, out_band, map_fun,
                map_args=None, map_kwargs=None):
    if isinstance(in_bands, Iterable):
        block_write_multi(in_ds, in_bands, out_band, map_fun,
                          map_args=map_args, map_kwargs=map_kwargs)
    else:
        block_write_one(in_ds, in_bands, out_band, map_fun,
                        map_args=map_args, map_kwargs=map_kwargs)


def block_write_one(in_ds, in_band, out_band, map_fun,
                    map_args=None, map_kwargs=None):
    block_xsize, block_ysize = in_band.GetBlockSize()
    for b_y, yoff in enumerate(range(0, in_ds.RasterYSize, block_ysize)):
        for b_x, xoff in enumerate(range(0, in_ds.RasterXSize, block_xsize)):
            win_xsize, win_ysize = in_band.GetActualBlockSize(b_x, b_y)

            in_data = in_band.ReadAsArray(
                xoff=xoff, yoff=yoff, win_xsize=win_xsize, win_ysize=win_ysize)
            in_data = np.ma.masked_array(in_data,
                                         mask=in_data == in_band.GetNoDataValue(),
                                         fill_value=out_band.GetNoDataValue())

            if map_args and map_kwargs:
                out_data = map_fun(in_data, *map_args, **map_kwargs)
            elif map_args:
                out_data = map_fun(in_data, *map_args)
            elif map_kwargs:
                out_data = map_fun(in_data, **map_kwargs)
            else:
                out_data = map_fun(in_data)

            out_band.WriteArray(out_data, xoff=xoff, yoff=yoff)


def block_write_multi(in_ds, in_bands, out_band, map_fun,
                      map_args=None, map_kwargs=None):
    first_band = in_bands[0]
    block_xsize, block_ysize = first_band.GetBlockSize()
    for b_y, yoff in enumerate(range(0, in_ds.RasterYSize, block_ysize)):
        for b_x, xoff in enumerate(range(0, in_ds.RasterXSize, block_xsize)):
            win_xsize, win_ysize = first_band.GetActualBlockSize(b_x, b_y)
            no_data = out_band.GetNoDataValue()

            in_datas = []
            for in_band in in_bands:
                in_data = in_band.ReadAsArray(
                    xoff=xoff, yoff=yoff, win_xsize=win_xsize, win_ysize=win_ysize)
                in_data = np.ma.masked_array(in_data,
                                             mask=in_data == in_band.GetNoDataValue(),
                                             fill_value=no_data)
                in_datas.append(in_data)

            if map_args and map_kwargs:
                out_data = map_fun(in_datas, *map_args, **map_kwargs)
            elif map_args:
                out_data = map_fun(in_datas, *map_args)
            elif map_kwargs:
                out_data = map_fun(in_datas, **map_kwargs)
            else:
                out_data = map_fun(in_datas)

            out_band.WriteArray(out_data, xoff=xoff, yoff=yoff)



import os
from osgeo import gdal, osr
from .transform import ds_name, context_file
from ._const import CREATION, CONFIG


__all__ = ['grib_to_tif']


def grib_to_tif(ds, out_path=None, **kwargs):
    ds, ras = ds_name(ds)

    if os.path.splitext(os.path.basename(ras))[1] != '.grib':
        return

    if out_path:
        out_file = context_file(ras, out_path)
    else:
        out_file = os.path.join(os.path.dirname(ras), os.path.splitext(
            os.path.basename(ras))[0] + '.tif')

    if os.path.exists(out_file):
        return out_file

    proj = "+proj=longlat +datum=WGS84 +ellps=WGS84"
    SpatialRef = osr.SpatialReference()
    SpatialRef.ImportFromProj4(proj)
    srs = kwargs.pop('dstSRS', SpatialRef)
    option = gdal.WarpOptions(multithread=True, options=CONFIG,
                              creationOptions=CREATION, **kwargs,
                              dstSRS=srs)
    gdal.Warp(out_file, ds, options=option)

    return out_file



import os
from osgeo import gdal
from .transform import ds_name, context_file
from ._const import CREATION, CONFIG

__all__ = ['mosaic']


def mosaic(ras_paths, out_path, **kwargs):
    ds = ras_paths[0]
    ds, ras = ds_name(ds)
    out_file = context_file(ras, out_path)

    if os.path.exists(out_file):
        return out_file

    separate = kwargs.pop('separate', False)
    resample_alg = kwargs.pop('resampleAlg', gdal.GRA_Average)
    gdal.BuildVRT('/vsimem/Mosaic.vrt', ras_paths, separate=separate)
    ds = gdal.Open('/vsimem/Mosaic.vrt')
    option = gdal.WarpOptions(multithread=True, options=CONFIG,
                              creationOptions=CREATION, **kwargs,
                              resampleAlg=resample_alg)
    gdal.Warp(out_file, ds, options=option)

    return out_file



import os
from osgeo import gdal, ogr, osr

__all__ = ['proj_shapefile']


def proj_shapefile(in_shp, out_shp, in_proj=None,
                   out_proj="+proj=longlat +datum=WGS84 +ellps=WGS84"):
    gdal.SetConfigOption("SHAPE_ENCODING", 'utf-8')

    # Filename of input OGR file
    driver = ogr.GetDriverByName("ESRI Shapefile")
    if isinstance(in_shp, ogr.Layer):
        inLayer = in_shp
    else:
        source_ds = driver.Open(in_shp)
        inLayer = source_ds.GetLayer()

    # input SpatialReference
    inSpatialRef = inLayer.GetSpatialRef()
    if not inSpatialRef:
        if in_proj is not None:
            if isinstance(in_proj, osr.SpatialReference):
                inSpatialRef = in_proj
            else:
                inSpatialRef = osr.SpatialReference()
                inSpatialRef.ImportFromProj4(in_proj)
        else:
            raise(ValueError("in_proj must be set"))

    # output SpatialReference
    if isinstance(out_proj, osr.SpatialReference):
        outSpatialRef = out_proj
    elif os.path.isfile(out_proj):
        ds = driver.Open(out_proj)
        outSpatialRef = ds.GetLayer().GetSpatialRef()
        ds = None
    else:
        outSpatialRef = osr.SpatialReference()
        outSpatialRef.ImportFromProj4(out_proj)

    # create the CoordinateTransformation
    if int(gdal.__version__[0]) >= 3:
        outSpatialRef.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
    coordTrans = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)

    # create the output layer
    if 'vsimem' not in os.path.dirname(out_shp):
        if (not os.path.exists(os.path.dirname(out_shp)) and
                os.path.dirname(out_shp) != ''):
            os.makedirs(os.path.dirname(out_shp))
    if os.path.exists(out_shp):
        driver.DeleteDataSource(out_shp)
    outDataSet = driver.CreateDataSource(out_shp)
    outLayer = outDataSet.CreateLayer(out_shp, outSpatialRef)

    # add fields
    inLayerDefn = inLayer.GetLayerDefn()
    for i in range(0, inLayerDefn.GetFieldCount()):
        fieldDefn = inLayerDefn.GetFieldDefn(i)
        outLayer.CreateField(fieldDefn)

    # get the output layer's feature definition
    outLayerDefn = outLayer.GetLayerDefn()

    # loop through the input features
    inFeature = inLayer.GetNextFeature()
    while inFeature:
        # get the input geometry
        geom = inFeature.GetGeometryRef()
        # reproject the geometry
        geom.Transform(coordTrans)
        # create a new feature
        outFeature = ogr.Feature(outLayerDefn)
        # set the geometry and attribute
        outFeature.SetGeometry(geom)
        for i in range(0, outLayerDefn.GetFieldCount()):
            outFeature.SetField(outLayerDefn.GetFieldDefn(
                i).GetNameRef(), inFeature.GetField(i))
        # add the feature to the shapefile
        outLayer.CreateFeature(outFeature)
        # dereference the features and get the next input feature
        outFeature = None
        inFeature = inLayer.GetNextFeature()

    source_ds = None
    inLayer = None
    outLayer = None
    outDataSet = None

    return out_shp



import os
from osgeo import gdal, ogr
from ._const import CREATION, CONFIG


__all__ = ['shp_to_raster']


def shp_to_raster(shp, attr, out_path, ds_eg, tem_path, **kwargs):
    # create out put name
    out_file = os.path.join(out_path, os.path.splitext(
        os.path.basename(shp))[0] + '.tif')
    if os.path.exists(out_file):
        return
    tem_file = os.path.join(tem_path, os.path.splitext(
        os.path.basename(shp))[0] + '.tif')

    # extent warp options
    ds_ex = gdal.Translate('/vsimem/_extent.tif', ds_eg, bandList=[1])
    t = ds_eg.GetGeoTransform()
    temp_option = gdal.WarpOptions(multithread=True, options=CONFIG,
                                   creationOptions=CREATION, **kwargs,
                                   xRes=t[1] / 10, yRes=t[5] / 10,
                                   outputType=gdal.GDT_Float64)

    ds_tem = gdal.Warp(tem_file, ds_ex, options=temp_option)
    band = ds_tem.GetRasterBand(1)
    option = gdal.WarpOptions(multithread=True, options=CONFIG,
                              creationOptions=CREATION,  **kwargs,
                              xRes=t[1], yRes=t[5], resampleAlg=gdal.GRA_Average,
                              outputType=gdal.GDT_Float64)

    driver = ogr.GetDriverByName('ESRI Shapefile')
    shp_factor = driver.Open(shp)
    layer = shp_factor.GetLayer()

    # create and use RasterizeLayer
    band.Fill(band.GetNoDataValue())
    gdal.RasterizeLayer(ds_tem, [1], layer,
                        options=["ATTRIBUTE=%s" % attr, 'ALL_TOUCHED=TRUE'])
    ds_out = gdal.Warp(out_file, ds_tem, options=option)

    # deal with units
    if os.path.splitext(os.path.basename(out_file))[0] == 'permeability':
        band_out = ds_out.GetRasterBand(1)
        no_data = band_out.GetNoDataValue()
        values = ds_out.ReadAsArray()
        values[values != no_data] = values[values != no_data] / 100
        band_out.WriteArray(values)

        band = ds_tem.GetRasterBand(1)
        no_data = band.GetNoDataValue()
        values = ds_tem.ReadAsArray()
        values[values != no_data] = values[values != no_data] / 100
        band.WriteArray(values)

    band_out = None
    ds_out = None
    band = None
    ds_tem = None
    shp_factor = None
    layer = None



import os
from osgeo import ogr, osr

__all__ = ['shp_buffer']


def shp_buffer(in_shp, out_shp, buffdist, in_proj=None):
    # Filename of input OGR file
    driver = ogr.GetDriverByName("ESRI Shapefile")
    if isinstance(in_shp, ogr.Layer):
        inLayer = in_shp
    else:
        source_ds = driver.Open(in_shp)
        inLayer = source_ds.GetLayer()

    # input SpatialReference
    inSpatialRef = inLayer.GetSpatialRef()
    if not inSpatialRef:
        if in_proj is not None:
            if isinstance(in_proj, osr.SpatialReference):
                inSpatialRef = in_proj
            else:
                inSpatialRef = osr.SpatialReference()
                inSpatialRef.ImportFromProj4(in_proj)
        else:
            raise(ValueError("in_proj must be set"))

    # create the output layer
    if 'vsimem' not in os.path.dirname(out_shp):
        if (not os.path.exists(os.path.dirname(out_shp)) and
                os.path.dirname(out_shp) != ''):
            os.makedirs(os.path.dirname(out_shp))
    if os.path.exists(out_shp):
        driver.DeleteDataSource(out_shp)
    outDataSet = driver.CreateDataSource(out_shp)
    outLayer = outDataSet.CreateLayer(out_shp, inSpatialRef)

    # add fields
    inLayerDefn = inLayer.GetLayerDefn()
    for i in range(0, inLayerDefn.GetFieldCount()):
        fieldDefn = inLayerDefn.GetFieldDefn(i)
        outLayer.CreateField(fieldDefn)

    # get the output layer's feature definition
    outLayerDefn = outLayer.GetLayerDefn()

    # loop through the input features
    inFeature = inLayer.GetNextFeature()
    while inFeature:
        # get the input geometry
        geom = inFeature.GetGeometryRef()
        # create a new feature
        outFeature = ogr.Feature(outLayerDefn)
        # set the geometry and attribute
        outFeature.SetGeometry(geom.Buffer(float(buffdist)))
        for i in range(0, outLayerDefn.GetFieldCount()):
            outFeature.SetField(outLayerDefn.GetFieldDefn(
                i).GetNameRef(), inFeature.GetField(i))
        # add the feature to the shapefile
        outLayer.CreateFeature(outFeature)
        # dereference the features and get the next input feature
        outFeature = None
        inFeature = inLayer.GetNextFeature()

    inLayer = None
    outLayer = None
    outDataSet = None
    source_ds = None

    return out_shp



import os
import numpy as np
from osgeo import gdal
from ._const import CREATION
from .block_write import block_write
from .transform import rep_file, ds_name

__all__ = ['convert_uint8']


def map_no_data(in_data):
    return in_data.filled()


def convert_uint8(ds, in_no_data=None, out_no_data=255):
    ds, ras = ds_name(ds)
    frist_band = ds.GetRasterBand(1)

    if (frist_band.DataType != gdal.GDT_Byte or
            frist_band.ReadAsArray(0, 0, 1, 1).dtype != np.int8):
        return ras

    if frist_band.GetNoDataValue() is not None:
        in_no_data = frist_band.GetNoDataValue()
    if in_no_data is None:
        raise(ValueError("in_no_data must be initialed"))

    option = gdal.WarpOptions(multithread=True,
                              creationOptions=CREATION,
                              outputType=gdal.GDT_Byte)
    out_file = rep_file(os.path.dirname(ras), ras)
    ds_out = gdal.Warp(out_file, ras, options=option)

    for i in range(1, 1 + ds.RasterCount):
        band = ds.GetRasterBand(i)
        band_out = ds_out.GetRasterBand(i)
        if band.GetNoDataValue() is None:
            band.SetNoDataValue(in_no_data)
        band_out.SetNoDataValue(out_no_data)
        block_write(ds, band, band_out, map_no_data)

    ds = None
    ds_out = None
    gdal.GetDriverByName('GTiff').Delete(ras)
    os.rename(out_file, ras)

    return ras



import os
import locale
from osgeo import gdal, ogr, osr


__all__ = ['split_shp']


def split_shp(outDir, shp_paths):
    if not os.path.exists(outDir):
        os.mkdir(outDir)
    gdal.SetConfigOption("GDAL_FILENAME_IS_UTF8", "YES")
    gdal.SetConfigOption("SHAPE_ENCODING", locale.getpreferredencoding())
    for shp in shp_paths:
        # Filename of input OGR file
        driver = ogr.GetDriverByName("ESRI Shapefile")
        source_ds = driver.Open(shp)
        inLayer = source_ds.GetLayer()

        # input SpatialReference
        inSpatialRef = inLayer.GetSpatialRef()

        # output SpatialReference
        outSpatialRef = osr.SpatialReference()
        prj_srs = "+proj=longlat +datum=WGS84 +ellps=WGS84"
        outSpatialRef.ImportFromProj4(prj_srs)

        # create the CoordinateTransformation
        if int(gdal.__version__[0]) >= 3:
            outSpatialRef.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        coordTrans = osr.CoordinateTransformation(inSpatialRef, outSpatialRef)

        # loop through the input features
        inLayerDefn = inLayer.GetLayerDefn()
        inFeature = inLayer.GetNextFeature()
        while inFeature:
            # create the output layer
            outFile = outDir + '/' + inFeature.GetField(0) + '.shp'
            if os.path.exists(outFile):
                driver.DeleteDataSource(outFile)
            outDataSet = driver.CreateDataSource(outFile)
            outLayer = outDataSet.CreateLayer(outFile, outSpatialRef)
            # get the output layer's feature definition
            outLayerDefn = outLayer.GetLayerDefn()

            # add fields
            for i in range(0, inLayerDefn.GetFieldCount()):
                fieldDefn = inLayerDefn.GetFieldDefn(i)
                outLayer.CreateField(fieldDefn)

            # get the input geometry
            geom = inFeature.GetGeometryRef()
            # reproject the geometry
            geom.Transform(coordTrans)
            # create a new feature
            outFeature = ogr.Feature(outLayerDefn)
            # set the geometry and attribute
            outFeature.SetGeometry(geom)
            for i in range(0, outLayerDefn.GetFieldCount()):
                outFeature.SetField(outLayerDefn.GetFieldDefn(
                    i).GetNameRef(), inFeature.GetField(i))
            # add the feature to the shapefile
            outLayer.CreateFeature(outFeature)
            # dereference the features and get the next input feature
            outFeature = None
            outDataSet = None
            outLayer = None
            inFeature = inLayer.GetNextFeature()

        outLayer = None
        inLayer = None
        source_ds = None
        outDataSet = None
